@@ -115,10 +115,8 @@ struct rpcbdump_short
 
 #ifdef PORTMAP
 static void ip_ping (u_short, char *, int, char **);
-static CLIENT *clnt_com_create (struct sockaddr_in *, u_long, u_long, int *,
-				char *);
 static void pmapdump (int, char **);
-static void get_inet_address (struct sockaddr_in *, char *);
+static CLIENT *ip_getclient(const char *hostname, rpcprog_t prognum, rpcvers_t versnum, const char *proto);
 #endif
 
 static bool_t reply_proc (void *, struct netbuf *, struct netconfig *);
@@ -356,38 +354,17 @@ local_rpcb (rpcprog_t prog, rpcvers_t vers)
 }
 
 #ifdef PORTMAP
-static CLIENT *
-clnt_com_create (addr, prog, vers, fdp, trans)
-     struct sockaddr_in *addr;
-     u_long prog;
-     u_long vers;
-     int *fdp;
-     char *trans;
+static enum clnt_stat
+ip_ping_one(client, vers)
+     CLIENT *client;
+     u_int32_t vers;
 {
-  CLIENT *clnt;
+  struct timeval to = { .tv_sec = 10, .tv_usec = 0 };
 
-  if (strcmp (trans, "tcp") == 0)
-    {
-      clnt = clnttcp_create (addr, prog, vers, fdp, 0, 0);
-    }
-  else
-    {
-      struct timeval to;
-
-      to.tv_sec = 5;
-      to.tv_usec = 0;
-      clnt = clntudp_create (addr, prog, vers, to, fdp);
-    }
-  if (clnt == (CLIENT *) NULL)
-    {
-      clnt_pcreateerror ("rpcinfo");
-      if (vers == MIN_VERS)
-	printf ("program %lu is not available\n", prog);
-      else
-	printf ("program %lu version %lu is not available\n", prog, vers);
-      exit (1);
-    }
-  return (clnt);
+  (void) CLNT_CONTROL (client, CLSET_VERS, &vers);
+  return CLNT_CALL (client, NULLPROC, (xdrproc_t) xdr_void,
+		    (char *) NULL, (xdrproc_t) xdr_void, (char *) NULL,
+		    to);
 }
 
 /*
@@ -398,17 +375,15 @@ clnt_com_create (addr, prog, vers, fdp, trans)
  * version 0 calls succeeds, it tries for MAXVERS call and repeats the same.
  */
 static void
-ip_ping (portnum, trans, argc, argv)
+ip_ping (portnum, proto, argc, argv)
      u_short portnum;
-     char *trans;
+     char *proto;
      int argc;
      char **argv;
 {
   CLIENT *client;
-  int fd = RPC_ANYFD;
-  struct timeval to;
-  struct sockaddr_in addr;
   enum clnt_stat rpc_stat;
+  const char *hostname;
   u_long prognum, vers, minvers, maxvers;
   struct rpc_err rpcerr;
   int failure = 0;
@@ -418,10 +393,9 @@ ip_ping (portnum, trans, argc, argv)
       usage ();
       exit (1);
     }
-  to.tv_sec = 10;
-  to.tv_usec = 0;
+
+  hostname = argv[0];
   prognum = getprognum (argv[1]);
-  get_inet_address (&addr, argv[0]);
   if (argc == 2)
     {				/* Version number not known */
       /*
@@ -434,11 +408,10 @@ ip_ping (portnum, trans, argc, argv)
     {
       vers = getvers (argv[2]);
     }
-  addr.sin_port = htons (portnum);
-  client = clnt_com_create (&addr, prognum, vers, &fd, trans);
-  rpc_stat = CLNT_CALL (client, NULLPROC, (xdrproc_t) xdr_void,
-			(char *) NULL, (xdrproc_t) xdr_void, (char *) NULL,
-			to);
+
+  client = ip_getclient(hostname, prognum, vers, proto);
+
+  rpc_stat = ip_ping_one(client, vers);
   if (argc != 2)
     {
       /* Version number was known */
@@ -447,8 +420,8 @@ ip_ping (portnum, trans, argc, argv)
       (void) CLNT_DESTROY (client);
       return;
     }
+
   /* Version number not known */
-  (void) CLNT_CONTROL (client, CLSET_FD_NCLOSE, (char *) NULL);
   if (rpc_stat == RPC_PROGVERSMISMATCH)
     {
       clnt_geterr (client, &rpcerr);
@@ -461,12 +434,7 @@ ip_ping (portnum, trans, argc, argv)
        * Oh dear, it DOES support version 0.
        * Let's try version MAX_VERS.
        */
-      (void) CLNT_DESTROY (client);
-      addr.sin_port = htons (portnum);
-      client = clnt_com_create (&addr, prognum, MAX_VERS, &fd, trans);
-      rpc_stat = CLNT_CALL (client, NULLPROC, (xdrproc_t) xdr_void,
-			    (char *) NULL, (xdrproc_t) xdr_void,
-			    (char *) NULL, to);
+      rpc_stat = ip_ping_one(client, MAX_VERS);
       if (rpc_stat == RPC_PROGVERSMISMATCH)
 	{
 	  clnt_geterr (client, &rpcerr);
@@ -495,21 +463,15 @@ ip_ping (portnum, trans, argc, argv)
       (void) pstatus (client, prognum, (u_long) 0);
       exit (1);
     }
-  (void) CLNT_DESTROY (client);
   for (vers = minvers; vers <= maxvers; vers++)
     {
-      addr.sin_port = htons (portnum);
-      client = clnt_com_create (&addr, prognum, vers, &fd, trans);
-      rpc_stat = CLNT_CALL (client, NULLPROC, (xdrproc_t) xdr_void,
-			    (char *) NULL, (xdrproc_t) xdr_void,
-			    (char *) NULL, to);
+      rpc_stat = ip_ping_one(client, vers);
       if (pstatus (client, prognum, vers) < 0)
 	failure = 1;
-      (void) CLNT_DESTROY (client);
     }
   if (failure)
     exit (1);
-  (void) close (fd);
+  (void) CLNT_DESTROY (client);
   return;
 }
 
@@ -521,9 +483,7 @@ pmapdump (argc, argv)
      int argc;
      char **argv;
 {
-  struct sockaddr_in server_addr;
   struct pmaplist *head = NULL;
-  int socket = RPC_ANYSOCK;
   struct timeval minutetimeout;
   register CLIENT *client;
   struct rpcent *rpc;
@@ -539,10 +499,13 @@ pmapdump (argc, argv)
   if (argc == 1)
     {
       host = argv[0];
-      get_inet_address (&server_addr, host);
-      server_addr.sin_port = htons (PMAPPORT);
-      client = clnttcp_create (&server_addr, PMAPPROG, PMAPVERS,
-			       &socket, 50, 500);
+
+      /* This is a little bit more complicated than it should be.
+       * ip_getclient will do an rpcb_getaddr call to identify the
+       * port of the portmapper - but it works, and it's easier than
+       * creating a copy of ip_getclient that avoids the getaddr call.
+       */
+      client = ip_getclient(host, PMAPPROG, PMAPVERS, "tcp");
     }
   else
     client = local_rpcb (PMAPPROG, PMAPVERS);
@@ -609,48 +572,74 @@ pmapdump (argc, argv)
     }
 }
 
-static void
-get_inet_address (addr, host)
-     struct sockaddr_in *addr;
-     char *host;
+/*
+ * Try to obtain the address of a given host/program/version, using the
+ * specified protocol (one of udp or tcp).
+ * This loops over all netconfig entries (according to the order given by
+ * netpath and the config file), and tries to resolve the hostname, and obtain
+ * the address using rpcb_getaddr.
+ */
+CLIENT *
+ip_getclient(hostname, prognum, versnum, proto)
+     const char *hostname;
+     rpcprog_t prognum;
+     rpcvers_t versnum;
+     const char *proto;
 {
-  struct netconfig *nconf;
-  struct addrinfo hints, *res;
-  int error;
+  void *handle;
+  enum clnt_stat saved_stat = RPC_SUCCESS;
+  struct netconfig *nconf, *result = NULL;
+  struct netbuf bind_address;
+  struct sockaddr_storage __sa;
+  CLIENT *client;
 
-  (void) memset ((char *) addr, 0, sizeof (*addr));
-  addr->sin_addr.s_addr = inet_addr (host);
-  if (addr->sin_addr.s_addr == -1 || addr->sin_addr.s_addr == 0)
+  memset(&bind_address, 0, sizeof(bind_address));
+  bind_address.maxlen = sizeof(__sa);
+  bind_address.buf = &__sa;
+
+  handle = setnetconfig();
+  while ((nconf = getnetconfig(handle)) != NULL)
     {
-      if ((nconf = __rpc_getconfip ("udp")) == NULL &&
-	  (nconf = __rpc_getconfip ("tcp")) == NULL)
-	{
-	  fprintf (stderr, "rpcinfo: couldn't find a suitable transport\n");
-	  exit (1);
-	}
+      if (!strcmp(nconf->nc_proto, proto)) {
+        if (rpcb_getaddr(prognum, versnum, nconf, &bind_address, hostname))
+          {
+	    result = getnetconfigent(nconf->nc_netid);
+	    endnetconfig(handle);
+	    break;
+	  }
+
+        if (rpc_createerr.cf_stat != RPC_UNKNOWNHOST)
+          {
+            clnt_pcreateerror (hostname);
+	    exit (1);
+	  }
+
+	saved_stat = rpc_createerr.cf_stat;
+      }
+    }
+
+  if (result == NULL)
+    {
+      if (saved_stat != RPC_SUCCESS)
+        {
+          rpc_createerr.cf_stat = saved_stat;
+          clnt_pcreateerror (hostname);
+        }
       else
-	{
-	  memset (&hints, 0, sizeof hints);
-	  hints.ai_family = AF_INET;
-	  if ((error = getaddrinfo (host, "rpcbind", &hints, &res)) != 0 &&
-              (error = getaddrinfo (host, "portmapper", &hints, &res)) != 0)
-	    {
-	      fprintf (stderr, "rpcinfo: %s: %s\n",
-		       host, gai_strerror (error));
-	      exit (1);
-	    }
-	  else
-	    {
-	      memcpy (addr, res->ai_addr, res->ai_addrlen);
-	      freeaddrinfo (res);
-	    }
-	  (void) freenetconfigent (nconf);
-	}
+        fprintf (stderr, "Cannot find suitable transport for protocol %s\n", proto);
+
+      exit (1);
     }
-  else
+
+  client = clnt_tli_create(RPC_ANYFD, result, &bind_address, prognum, versnum, 0, 0);
+  if (client == NULL)
     {
-      addr->sin_family = AF_INET;
+      clnt_pcreateerror(hostname);
+      exit (1);
     }
+
+  freenetconfigent(result);
+  return client;
 }
 #endif /* PORTMAP */
 
